<?php
namespace Controller;

use Zeedhi\Framework\DataSource\Manager;
use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\DTO\Response;
use Zeedhi\Framework\DataSource\DataSet;
use Util\Messages;

class Priority extends \Generated\Controller\Priority {

    /** @var  \Service\Priority $priorityService */
    protected $priorityService;
    /** @var \Util\Utilities $utilities */
    protected $utilities;
    /** @var  \Service\Task $taskService */
    protected $taskService;

    public function __construct(Manager $dataSourceManager, \Util\Utilities $utilities, \Service\Priority $priorityService, \Service\Task $taskService){
        $this->utilities = $utilities;
        $this->priorityService = $priorityService;
        $this->taskService = $taskService;
        parent::__construct($dataSourceManager);
    }

    /**
     * Rewriting our find() method found on Crud,
     * inserting a condition to criteria.
     *
     * @param Request\Filter $request
     * @param Response $response
     */
    public function find(Request\Filter $request, Response $response){
        $request->getFilterCriteria()->addCondition("ORGANIZATION", "=", $this->utilities->getCurrentOrganizationId());
        parent::find($request, $response);
    }

    /**
     * Rewriting our delete() method found on Crud.
     *
     * @param Request\DataSet $request
     * @param Response $response
     */
    public function delete(Request $request, Response $response, $showMessage = false)
    {
        try {
            parent::delete($request, $response, $showMessage);
            $response->addNotification(new Response\Notification(Messages::SUC_SUCCESSFUL_REMOVED));
        } catch (\Exception $ex) {
            $errorMessage = $response->getError()->getMessage();
            $isUsed = strpos($errorMessage, "ORA-02292");
            $priorityId = $request->getDataSet()->getRows()[0]["ID"];
            $message = '';
            $message .= Messages::ERR_NOT_POSSIBLE_TO_DELETE_ISSUE_PRIORITY_ALREADY_USED;
            $coma = '<br/>';
            if($isUsed) {
                $taskPriorities = $this->taskService->getTasksByPriority($priorityId);
                $timesUsed = count($taskPriorities);
                $message .= $timesUsed;
                $this->utilities->treatResponseError($response, $message." issues.");
            }else{
                $this->utilities->treatResponseError($response, $errorMessage);
            }
            $this->utilities->treatResponseError($response, 'Não foi possível excluir a prioridade, pois ela está sendo usada em outro lugar.');
        }
    }

    /**
     * Called when try to save a priority, whenever
     * if user is trying to update an existent or
     * create a new one.
     *
     * @param Request\DataSet $request
     * @param Response $response
     */
    public function updateIssuePriorities(Request\DataSet $request, Response $response){
        try {
            $parameter = $request->getDataSet()->getRows()[0];
            $priorityId     = $parameter['ID'] ? $parameter['ID'] : null;
            $priorityLabel  = $parameter['LABEL'] ? $parameter['LABEL'] : null;
            $priorityColor  = $parameter['COLOR'] ? $parameter['COLOR'] : null;
            if(empty($priorityId)){
                $dataSet = $this->priorityService->createIssuePriorities($priorityLabel, $priorityColor);
            } else {
                $dataSet = $this->priorityService->updateIssuePriorities($priorityId, $priorityLabel, $priorityColor);
            }
            $response->addDataSet(new DataSet('issuePriorities', $dataSet));
            $response->addNotification(new Response\Notification(\Util\Messages::SUC_SUCCESSFUL_SAVED));
        } catch(\Exception $e){
            $response->addMessage(new Response\Message(\Util\Messages::ERR_NOT_SUCCESSFUL_SAVED_EXPLANATION.$e->getMessage()));
        }
    }

}