<?php

namespace Factory;
use Zeedhi\DevOps\Model\Projectmanagement\Entities;

/**
 * Created by: Stanley Sathler
 * This Factory is responsible by the management of
 * Task Options screen.
 *
 * Class TaskOptions
 * @package Factory
 */

class TaskOptions {

	private $entityManager;

	public function __construct(\Doctrine\ORM\EntityManager $entityManager){
		$this->entityManager = $entityManager;
	}

    /**
     * Creates a new task status.
     * @param $label
     * @param string $isIssue
     * @param $control
     * @param Entities\TaskType $taskType
     * @param null $isTask
     * @param null $usedInTask
     * @param null $usedInIssue
     * @return Entities\TaskStatus
     */
	public function createStatus($label,$isIssue = 'N',$control = null,
                                 Entities\TaskType $taskType = null,$isTask = null,$usedInTask = null,$usedInIssue = null){
		$entity = new Entities\TaskStatus();
		$entity->setControl($control);
		$entity->setLabel($label);
		$entity->setIsIssue($isIssue);
		$entity->setTaskType($taskType);
		$entity->setIsTask($isTask);
		$entity->setUsedInTask($usedInTask);
		$entity->setUsedInIssue($usedInIssue);
		return $entity;
	}

    /**
     * Updates an existent status.
     * @param Entities\TaskStatus $entity
     * @param $label
     * @param string $isIssue
     * @param null $control
     * @param Entities\TaskType $taskType
     * @param $isTask String
     * @param null $usedInTask
     * @param null $usedInIssue
     * @return Entities\TaskStatus
     */
	public function updateStatus(Entities\TaskStatus $entity, $label,$isIssue = 'N',$control = null,
                                 Entities\TaskType $taskType = null,$isTask = 'N',$usedInTask = null,$usedInIssue = null){
		$entity->setControl($control);
		$entity->setLabel($label);
		$entity->setIsIssue($isIssue);
		$entity->setTaskType($taskType);
        $entity->setIsTask($isTask);
        $entity->setUsedInTask($usedInTask);
        $entity->setUsedInIssue($usedInIssue);
		return $entity;
	}

	public function updateTaskTypeFromTaskStatus(Entities\TaskType $entityTaskType, $entityTaskStatus = null){
		$entityTaskType->setInitialTaskStatus($entityTaskStatus);
		return $entityTaskType;
	}

	public function createTaskOptionsParameter(Entities\TaskStatus $initialTaskStatus,
											   Entities\TaskStatus $finalTaskStatus,
											   $control = null, $label = null,
											   $taskType = null, $isIssue = 'N'){
		/** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskProperties */
		$entity = new Entities\TaskProperties();
		$entity->setControl($control);
		$entity->setLabel($label);
		$entity->setTaskType($taskType);
		$entity->setIsIssue($isIssue);
		$entity->setInitialTaskStatus($initialTaskStatus);
		$entity->setFinalTaskStatus($finalTaskStatus);
		return $entity;
	}

} 