<?php
namespace Service;

use Doctrine\ORM\EntityManager;
use React\Tests\Stream\UtilTest;
use Service\Calendar\TaskCalendar;
use Service\Calendar\ResourceCalendar;
use Util\InstanceProvider;
use Util\Messages;
use Util\Utilities;
use \Zeedhi\DevOps\Model\Projectmanagement\Repositories;
use \Zeedhi\DevOps\Model\Projectmanagement\Entities;
use \Zeedhi\DevOps\Model\Configurationmanager\Repositories as ConfigurationRepositories;
use \Zeedhi\DevOps\Model\Configurationmanager\Entities as ConfigurationEntities;



/**
 * Class Appropriation
 *
 * @package Service
 */
class Appropriation implements AppropriationInterface
{

    /** @var EntityManager $entityManager */
    protected $entityManager;
    /** @var \Util\Utilities $environment */
    protected $environment;
    /** @var \Factory\Appropriation $appropriation */
    protected $appropriationFactory;
    /** @var \Factory\Branch $branchFactory */
    protected $branchFactory;
    /** @var \Factory\RepositoryBranch $repositoryBranchFactory */
    protected $repositoryBranchFactory;
    /** @var \Factory\FctVendedororca $fctVendedororcaFactory */
    protected $fctVendedororcaFactory;
    /** @var Repositories\Appropriation $appropriationRepository */
    protected $appropriationRepository;
    /** @var Repositories\Allocation $allocationRepository  */
    protected $allocationRepository;
    /** @var Repositories\Directory $directoryRepository  */
    protected $directoryRepository;
    /** @var Structure */
    private $structureService;
    /** @var TaskCalendar */
    private $taskCalendar;
    /** @var ResourceCalendar */
    private $resourceCalendar;
    /** @var Repositories\TaskStatus $taskStatusRepository  */
    protected $taskStatusRepository;
    /** @var Repositories\TaskBranches $taskBranchRepository */
    protected $taskBranchesRepository;
    /** @var Repositories\ChangedFiles $changedFilesRepository */
    protected $changedFilesRepository;
    /** @var Repositories\Repository $repositoryRepository */
    protected $repositoryRepository;
    /** @var Repositories\TaskType $taskTypeRepository */
    protected $taskTypeRepository;
    /** @var Repositories\Task $taskRepository */
    protected $taskRepository;
    /** @var Repositories\Task $projectRepository */
    protected $projectRepository;
    /** @var Repositories\Resource $resourceRepository */
    protected $resourceRepository;
    /** @var ConfigurationRepositories\Subrelease $subreleaseRepository */
    protected $subreleaseRepository;
    /** @var ConfigurationRepositories\Subrelease branchRepository */
    protected $branchRepository;
    /** @var  Repositories\TaskRelationship $taskRelationshipRepository */
    protected $taskRelationshipRepository;
    protected $logDirFile;
    private $cdClienteTek;

    const SUBRELEASE_STATUS_IN_TEST = 'T';
    const SUBRELEASE_STATUS_PENDING = 'P';
    const CONTROL_BACKLOG = 'B';
    const USED_IN_ISSUE_YES = 'Y';
    const MAX_APPROPRIATION_PERCENT = 100;

    /**
     * @param EntityManager $entityManager
     * @param Utilities $environment
     * @param \Factory\Appropriation $appropriationFactory
     * @param Structure $structureService
     * @param TaskCalendar $taskCalendar
     * @param ResourceCalendar $resourceCalendar
     * @param \Factory\Branch $branchFactory
     * @param \Factory\RepositoryBranch $repositoryBranchFactory
     * @param \Factory\FctVendedororca $fctVendedororcaFactory
     * @param $logDirFile
     */
    public function __construct(EntityManager $entityManager, \Util\Utilities $environment, \Factory\Appropriation $appropriationFactory
        , Structure $structureService, TaskCalendar $taskCalendar, ResourceCalendar $resourceCalendar, \Factory\Branch $branchFactory
        , \Factory\RepositoryBranch $repositoryBranchFactory, \Factory\FctVendedororca $fctVendedororcaFactory, $logDirFile , $cdClienteTek)
    {
        $this->entityManager = $entityManager;
        $this->environment = $environment;
        $this->structureService = $structureService;
        $this->taskCalendar = $taskCalendar;
        $this->resourceCalendar = $resourceCalendar;
        $this->branchFactory           = $branchFactory;
        $this->appropriationFactory    = $appropriationFactory;
        $this->repositoryBranchFactory = $repositoryBranchFactory;
        $this->fctVendedororcaFactory = $fctVendedororcaFactory;
        $this->logDirFile = $logDirFile;
         $this->cdClienteTek = $cdClienteTek;
        $this->subreleaseRepository       = $this->entityManager->getRepository(Utilities::CONFIGURATION_ENTITY_DIR . 'Subrelease');
        $this->branchRepository           = $this->entityManager->getRepository(Utilities::CONFIGURATION_ENTITY_DIR . 'Branch');
        $this->appropriationRepository    = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Appropriation');
        $this->taskBranchesRepository     = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskBranches');
        $this->changedFilesRepository     = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ChangedFiles');
        $this->taskStatusRepository       = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskStatus');
        $this->repositoryRepository       = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Repository');
        $this->allocationRepository       = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Allocation');
        $this->taskTypeRepository         = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskType');
        $this->taskRepository             = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Task');
        $this->directoryRepository        = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Directory');
        $this->resourceRepository         = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Resource');
        $this->taskRelationshipRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskRelationship');
        $this->projectRepository          = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Project');
    }

    public function getAppropriation($appropriationId)
    {
        $appropriations = $this->appropriationRepository->getAppropriationData($appropriationId);
        foreach ($appropriations as &$appropriation) {
            $appropriation['PERCENT_COMPLETED'] = intval($appropriation['PERCENT_COMPLETED']);
        }
        return $appropriations;
    }

    public function findAllAppropriationsByDate($filters = array())
    {
        $filtersString = $this->environment->createQueryFiltered($filters);
        $appropriations = $this->appropriationRepository->getAllAppropriationsByDate($this->environment->getCurrentOrganizationId(), $filtersString);
        if (!empty($appropriations)) {
            $resourceId = $appropriations[0]['RESOURCE_T'];
            $day = $appropriations[0]['DAY'];
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource $resource*/
            $resource = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource', $resourceId);

            $dayHours = $this->resourceCalendar->getDatesCalendar($resource, $day, $day);
            $dayHours = $dayHours[$day];

            /** @var Repositories\WorkedTime $repositoryWorkTime */
            $repositoryWorkTime = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\WorkedTime');
            $workedTime = $repositoryWorkTime->getWorkTimeByUserAndDate($resourceId, $day);
            $totalAppropriation = 0;
            foreach ($appropriations as &$appropriation) {
                $totalAppropriation += $appropriation['TOTAL_APPROPRIATION'];
                $appropriation['TOTAL_APPROPRIATION'] = $this->environment->secondsToTime(intval($appropriation['TOTAL_APPROPRIATION']));
                if (empty($appropriation['TOTAL_APPROPRIATION'])) {
                    $appropriation['TOTAL_APPROPRIATION'] = '0';
                }
            }
            $totalAppropriationLine = array();
            $diffAppropriationLine = array();
            if (empty($workedTime)) {
                $diff = $dayHours - intval($totalAppropriation);
            } else {
                $diff = intval($workedTime[0]['WORK_TIME']) - intval($totalAppropriation);
            }
            $diffAppropriationLine['HOURS_CONTROL'] = $diff;
            if ($diff < 0) {
                $diffFront = '+';
            } elseif ($diff == 0) {
                $diffFront = '0';
            } else {
                $diffFront = '-';
            }
            if ($totalAppropriation == 0) {
                $totalAppropriationFront = '0';
            } else {
                $totalAppropriationFront = '';
            }
            $diffAppropriationLine['TOTAL_APPROPRIATION'] = $diffFront . $this->environment->secondsToTime(abs($diff));
            $totalAppropriationLine['TOTAL_APPROPRIATION'] = $totalAppropriationFront . $this->environment->secondsToTime(intval($totalAppropriation));

            $diffAppropriationLine['TASK_NAME'] = 'Diferença de Apropriação';
            $totalAppropriationLine['TASK_NAME'] = 'Total apropriado';
            $totalAppropriationLine['ORDER_FRONT'] = 1;
            $diffAppropriationLine['ORDER_FRONT'] = 2;
            $totalAppropriationLine['BOLDER'] = true;
            $diffAppropriationLine['BOLDER'] = true;

            array_push($appropriations, $totalAppropriationLine);
            array_push($appropriations, $diffAppropriationLine);
        }
        return $appropriations;
    }

    private function prepareAppropriationsGroupingByResourcesAndDates($appropriations){
        $appropriationsByResourceAndDates = array();
        foreach($appropriations as $appropriation) {
            if(empty($appropriationsByResourceAndDates[$appropriation['RESOURCE_T']])){
                $appropriationsByResourceAndDates[$appropriation['RESOURCE_T']] = array();
            }
            $appropriationsByResourceAndDates[$appropriation['RESOURCE_T']][$appropriation['DAY']] = $appropriation;
        }

        return $appropriationsByResourceAndDates;
    }

    private function unprepareAppropriationUngroupingResourcesAndDates($appropriationsGroupedByResourceAndDate){
        $appropriations = array();
        foreach($appropriationsGroupedByResourceAndDate as $resourceId => $dates){
            foreach($dates as $date)
                array_push($appropriations, $date);
        }
        return $appropriations;
    }

    private function buildEmptyTotalAppropriation($parent, $day, $dayHours){
        $appropriation = array(
            'PARENT' => ''.$parent,
            'DAY' => $day,
            'DAY_HOURS' => $dayHours,
            'RESOURCE_T' => ''.$parent,
            'TOTAL_APPROPRIATION' => '0',
            'ORDER_FRONT' => \DateTime::createFromFormat('d/m/Y', $day)->format('Ymd'),
            'DAY_FRONT' => ''.intval(\DateTime::createFromFormat('d/m/Y', $day)->format('w')+1)
        );

        return $appropriation;
    }

    /**
     * @param $appropriation
     * @param $diffAppropriationsMeanByStructure
     * @param $structureId
     * @param $structuresIdsParentsIds array($structureId => $parentId)
     * @return array
     */
    private function calculateDiffAppropriationsMeanByStructureOnDemand($appropriation, $structureId, $structuresIdsParentsIds, $diffAppropriationsMeanByStructure) {
        $totalDiff = $appropriation['HOURS_CONTROL'];
        $diffValue = $appropriation['DIFF_APPROPRIATION'];
        $nameColor = isset($appropriation['HOURS_CONTROL_NAME'])?$appropriation['HOURS_CONTROL_NAME']:null;
        if (!empty($structureId)) {
            if (empty($diffAppropriationsMeanByStructure[$structureId])){
                $diffAppropriationsMeanByStructure[$structureId] = array(
                    'valuePercent' => floatval($diffValue),
                    'value' => $totalDiff,
                    'length' => 1
                );
            } else {
                $oldValuePercent = $diffAppropriationsMeanByStructure[$structureId]['valuePercent'];
                $oldValue = $diffAppropriationsMeanByStructure[$structureId]['value'];
                $oldLength = $diffAppropriationsMeanByStructure[$structureId]['length'];
                $newValuePercent = $oldValuePercent + (floatval($diffValue) - $oldValuePercent) / ($oldLength + 1);
                $newValue = $oldValue + ($totalDiff - $oldValue) / ($oldLength + 1);
                $diffAppropriationsMeanByStructure[$structureId]['valuePercent'] = $newValuePercent;
                $diffAppropriationsMeanByStructure[$structureId]['value'] = $newValue;
                $diffAppropriationsMeanByStructure[$structureId]['length'] = $oldLength + 1;
            }
            if(isset($nameColor)){
                if(empty($diffAppropriationsMeanByStructure[$structureId]['HOURS_CONTROL_NAME'])){
                    $diffAppropriationsMeanByStructure[$structureId]['HOURS_CONTROL_NAME'] = $nameColor;
                }else if($diffAppropriationsMeanByStructure[$structureId]['HOURS_CONTROL_NAME'] != -1 && $nameColor == 1 ){
                    $diffAppropriationsMeanByStructure[$structureId]['HOURS_CONTROL_NAME'] = $nameColor;
                }
            }
            /* Recursive call to "bubble" average calculations to structure's parents. */
            $parentId = empty($structuresIdsParentsIds[$structureId]) ? null : $structuresIdsParentsIds[$structureId];
            $diffAppropriationsMeanByStructure = $this->calculateDiffAppropriationsMeanByStructureOnDemand($appropriation, $parentId, $structuresIdsParentsIds, $diffAppropriationsMeanByStructure);
        }
        return $diffAppropriationsMeanByStructure;
    }

    /**
     * @param $structures
     * @param $diffAppropriationsMeanByStructure
     * @param $structuresIds
     * @return array
     */
    private function buildStructuresAppropriationsRows($structures, $structuresIds, $diffAppropriationsMeanByStructure){
        foreach ($structures as $index => $structure) {
            // Avoid ID duplicates on tree data
            $structures[$index]['ID'] = 'STRUCT'.$structures[$index]['ID'];
            $structures[$index]['_hasChild'] = "true";
            if(empty($structures[$index]['PARENT'])){
                $structures[$index]['PARENT'] = empty($structures[$index]['PARENTID']) ? null : $structures[$index]['PARENTID'];
            }

            if(array_search($structures[$index]['PARENT'], $structuresIds) === false)
                $structures[$index]['PARENT'] = null;

            if (empty($diffAppropriationsMeanByStructure[$structures[$index]['ID']])) {
                $structures[$index]['DIFF_APPROPRIATION'] = '0%';
                $structures[$index]['HOURS_CONTROL'] = 0;
            } else {
                $structures[$index]['DIFF_APPROPRIATION'] = round($diffAppropriationsMeanByStructure[$structures[$index]['ID']]['valuePercent'], 2) . '%';
                $structures[$index]['HOURS_CONTROL'] = $diffAppropriationsMeanByStructure[$structures[$index]['ID']]['value'];
                //set name color on frontEnd
               if(isset($diffAppropriationsMeanByStructure[$structures[$index]['ID']]['HOURS_CONTROL_NAME'])){
                   $structures[$index]['HOURS_CONTROL_NAME'] = $diffAppropriationsMeanByStructure[$structures[$index]['ID']]['HOURS_CONTROL_NAME'];
               }

            }
            unSet($structures[$index]['PARENTID']);

            // Avoid ID duplicates on tree data
            if(!empty($structures[$index]['PARENT'])){
                $structures[$index]['PARENT'] = 'STRUCT'.$structures[$index]['PARENT'];
            }
        }
        return $structures;
    }

    /**
     * @param $filters
     * @return array
     */
    public function getFiltersAllAppropriations($filters)
    {
        $onDemand = true;
        $beginDate = null;
        $endDate = null;
        $resourceId = null;
        $filtersResources = array();
        $filtersStructure = array();
        foreach ($filters as $key => &$filter) {
            if ($filter['columnName'] == 'RESPONSIBLE') {
                $structures = $this->structureService->getStructuresOwnedByManagerOnline($filter['value']);
                $structureIds = array();
                foreach ($structures as $structure) {
                    array_push($structureIds, $structure['ID']);
                }
                $structureFilter = array(
                    'columnName' => 'STRUCTURE',
                    'value' => $structureIds,
                    'operator' => 'IN',
                );
                $filtersResources[] = array(
                    'columnName' => 'PER.STRUCTURE',
                    'value' => $structureIds,
                    'operator' => 'IN');
                $filtersStructure[] = array(
                    'columnName' => 'STR.ID',
                    'value' => $structureIds,
                    'operator' => 'IN');
                array_push($filters, $structureFilter);
                unSet($filters[$key]);
            }
            if ($filter['columnName'] == 'PARENT') {
                $onDemand = false;
                $resourceId = $filter['value'];
                $filter['columnName'] = 'ID';
            }

            if ($filter['columnName'] == 'PERIOD') {
                $beginDate = $filter['value'][0];
                $endDate = $filter['value'][1];
            }
            if($filter['columnName'] == 'STATUS') {
                $filtersResources[]  = array(
                    'columnName' => 'RES.STATUS',
                    'value' =>  $filter['value'],
                    'operator' => 'LIKE',
                );
            }
            if($filter['columnName'] == 'STRUCTURE') {
                $structures =  $this->structureService->getStructuresAllIdsStructureChildes($filter['value']);
                $filter['value'] = $structures;
                $filtersStructure[] = array(
                    'columnName' => 'STR.ID',
                    'value' =>  $structures,
                    'operator' => 'IN');
            }
            if($filter['columnName'] == 'RESOURCE_T') {
                $filtersStructure[] = array(
                    'columnName' => 'STR.ID',
                    'value' =>   $this->structureService->getStructuresAllIdsStructureParentByResourceIds($filter['value']),
                    'operator' => 'IN');
                $filtersResources[]  = array(
                    'columnName' => 'RES.ID',
                    'value' =>  $filter['value'],
                    'operator' => 'IN',
                );
            }
        }
        return array($onDemand, $beginDate, $endDate,  $filters, $resourceId,$filtersResources,$filtersStructure);
    }

    /**
     * @param Entities\ChangedFiles[] $changedFiles
     * @param $repositoryURL
     * @param null $restrictProjects
     * @return array
     */
    protected function getNameRepositoriesWithoutChangedFiles(array $changedFiles = array(), $repositoryURL,$restrictProjects = null, $relatedProjectIds = null)
    {

        $repositories = array();
        foreach ($changedFiles as $changedFile) {
            if (!in_array($changedFile->getRepository()->getUrl(), $repositoryURL) &&
                !in_array($changedFile->getRepository()->getName(), $repositories) &&
                in_array($changedFile->getRepository()->getProject()->getId(), $relatedProjectIds) &&
                (empty($restrictProjects) || in_array($changedFile->getRepository()->getProject()->getId(), $restrictProjects))
            ) {
                $repositories[] = $changedFile->getRepository()->getName();
            }
        }
        return $repositories;
    }

    /**
     * @param $taskId
     * @param $typeRelationship
     * @return array|null
     */
    protected function getRestrictProjectIdsForTaskMergeRequest($taskId, $typeRelationship)
    {
        $projects = null;
        if ($typeRelationship == Entities\TaskRelationship::MERGE_REQUEST_TYPE) {
            /** @var Entities\TaskRelationship $mergeRequestTaskRelationship */
            $mergeRequestTaskRelationship = $this->taskRelationshipRepository->findOneBy(array('task' => $taskId,'type'=>Entities\TaskRelationship::MERGE_REQUEST_TYPE));
            $relatedProject = $this->taskRelationshipRepository->findRelatedProjectIds($mergeRequestTaskRelationship->getTaskRelationship()->getId());
            if (empty($relatedProject)) {
                $projects = array($mergeRequestTaskRelationship->getTaskRelationship()->getProject()->getId());
            }else{
                $projects = explode(',',$relatedProject['RELATED_PROJECTS']);
                $projects[] = $relatedProject['MAIN_PROJECT'];
            }
        }
        return $projects;
    }

    /**
     * @param $issue
     * @param $taskTypeForFraming
     */
    public function setStatusFraming(Entities\TASK $issue, $taskTypeForFraming)
    {
        /** @var Entities\TaskStatus $taskStatus */
        if ($issue->getTaskStatus()
            && $issue->getTaskStatus()->getTaskType()
            && $issue->getTaskStatus()->getTaskType()->getId() == $taskTypeForFraming['ID']
        ) {
            $taskStatus = null;
        } else if ($taskTypeForFraming['INITIAL_TASK_STATUS']) {
            $taskStatus = $this->taskStatusRepository->find($taskTypeForFraming['INITIAL_TASK_STATUS']);
        } else {
            $taskStatus = $this->taskStatusRepository->findOneBy(array('taskType' => $taskTypeForFraming['ID']));
        }
        if ($taskStatus) {
            $issue->setTaskStatus($taskStatus);
        }
    }

    private function filterStructuresWithAppropriationChildren($structures, $appropriations) {
        $newStructures = array();
        $retrieveParentQueue = $appropriations;
        while(count($retrieveParentQueue)) {
            $appropriation = array_pop($retrieveParentQueue);

            if(!empty($appropriation['PARENT'])){
                for($i = 0, $n = count($structures), $foundStruct = false; !$foundStruct && $i < $n; $i++){
                    if(!empty($structures[$i]) && $structures[$i]['ID'] == $appropriation['PARENT']){
                        array_push($newStructures, $structures[$i]);
                        array_push($retrieveParentQueue, $structures[$i]);
                        unset($structures[$i]);
                        $foundStruct = true;
                    }
                }
            }
        }

        return $newStructures;
    }

    public function findAllAppropriations($filters = array())
    {
        list($onDemand, $beginDate, $endDate,  $filters, $resourceId,$filtersResources,$filtersStructure) =
            $this->getFiltersAllAppropriations($filters);

        if(!empty($resourceId) && strstr($resourceId, 'STRUCT') !== false){
            return array();
        }
        $filtersString = $this->environment->createQueryFiltered($filters);
        $filtersResourcesString = $this->environment->createQueryFiltered($filtersResources);
        $filtersStructureString = $this->environment->createQueryFiltered($filtersStructure);
        $appropriations = $this->appropriationRepository->getAllAppropriations($this->environment->getCurrentOrganizationId(),
            $filtersString, $onDemand,$filtersResourcesString);
        if (!$onDemand && !empty($resourceId)) {
            /** @var Repositories\WorkedTime $repositoryWorkTime */
            $repositoryWorkTime = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\WorkedTime');
            $appropriationsGroupedByResourcesAndDates = $this->prepareAppropriationsGroupingByResourcesAndDates($appropriations);
            if(empty($appropriationsGroupedByResourcesAndDates)){
                $appropriationsGroupedByResourcesAndDates = array($resourceId=>'');
            }
            foreach($appropriationsGroupedByResourcesAndDates as $resourceId => $dates) {
                /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource $resource */
                $resource = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource', $resourceId);
                $datesAndHours = $this->resourceCalendar->getDatesCalendar($resource, $beginDate, $endDate);
                foreach($datesAndHours as $date => $hour){
                    if(empty($dates[$date])){

                        if(empty($appropriationsGroupedByResourcesAndDates[$resourceId])){
                            $appropriationsGroupedByResourcesAndDates[$resourceId] =
                                array($date => $this->buildEmptyTotalAppropriation($resourceId, $date, $hour));
                        } else {
                            $appropriationsGroupedByResourcesAndDates[$resourceId][$date] =
                                $this->buildEmptyTotalAppropriation($resourceId, $date, $hour);
                        }
                    } else {
                        $appropriationsGroupedByResourcesAndDates[$resourceId][$date]['DAY_HOURS'] = $hour;
                    }
                }

                ksort($appropriationsGroupedByResourcesAndDates[$resourceId], SORT_STRING|SORT_DESC);
            }

            $appropriations = $this->unprepareAppropriationUngroupingResourcesAndDates($appropriationsGroupedByResourcesAndDates);

            foreach ($appropriations as &$appropriation) {
                $totalAppropriation = $appropriation['TOTAL_APPROPRIATION'];
                $dayHours = $appropriation['DAY_HOURS'];
                $workedTime = $repositoryWorkTime->getWorkTimeByUserAndDate($appropriation['RESOURCE_T'], $appropriation['DAY']);
                if (empty($workedTime)) {
                    $diff = $dayHours - intval($totalAppropriation);
                } else {
                    $diff = intval($workedTime[0]['WORK_TIME']) - intval($totalAppropriation);
                }
                $appropriation['HOURS_CONTROL'] = $diff;
                if ($diff < 0) {
                    $diffFront = '+';
                } elseif ($diff == 0) {
                    $diffFront = '0';
                } else {
                    $diffFront = '-';
                }
                if ($totalAppropriation == 0) {
                    $totalAppropriationFront = '0';
                } else {
                    $totalAppropriationFront = '';
                }
                $appropriation['DIFF_APPROPRIATION'] = $diffFront . $this->environment->secondsToTime(abs($diff));

                $appropriation['TOTAL_APPROPRIATION'] = $totalAppropriationFront . $this->environment->secondsToTime(intval($appropriation['TOTAL_APPROPRIATION']));
                $dayWeek = \Util\Utilities::getDayWeekPortugueseByNumber($appropriation['DAY_FRONT']);
                $appropriation['DAY_FRONT'] = $dayWeek . ' ' . $appropriation['DAY'];
            }
        } else {
            $structuresIds = array();
            $structuresIdsParentsIdsTree = array();
            $uniqueStructures = array();
            if (empty($structures)){
                $structures = $this->structureService->getStructureByFilter($filtersStructureString);
            }

            foreach($structures as $index => $structure) {
                $structId = $structure['ID'];
                $structures[$index]['ID'] = 'STRUCT'.$structId;

                if(array_search($structId, $structuresIds) === false){
                    array_push($uniqueStructures, $structure);
                    array_push($structuresIds, $structId);
                }
                if(!empty($structure['PARENT'])) {
                    $structures[$index]['PARENT'] = 'STRUCT'.$structure['PARENT'];
                    $structuresIdsParentsIdsTree['STRUCT'.$structure['ID']] = 'STRUCT'.$structure['PARENT'];
                } else if(!empty($structure['PARENTID'])){
                    $structures[$index]['PARENTID'] = 'STRUCT'.$structures[$index]['PARENTID'];
                    $structuresIdsParentsIdsTree['STRUCT'.$structure['ID']] = 'STRUCT'.$structure['PARENTID'];
                } else {
                    $structuresIdsParentsIdsTree['STRUCT'.$structure['ID']] = null;
                }
            }
            $structures = $uniqueStructures;

            $diffAppropriationsMeanByStructure = array();
            foreach ($appropriations as &$appropriation) {

                /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource $resource */
                $resource = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource', $appropriation['ID']);

                $totalAppropriationsDaily = explode(';', $appropriation['TOTAL_APPROPRIATION_DAILY']);
                if(empty($beginDate))
                    list($beginDate) = explode(',', $totalAppropriationsDaily[0]);
                if(empty($endDate))
                    list($beginDate) = explode(',', $totalAppropriationsDaily[count($totalAppropriationsDaily) - 1]);
                $totalAppropriationsDaily  = implode(';', $totalAppropriationsDaily);

                $datesAndHours = $this->resourceCalendar->getDatesCalendar($resource, $beginDate, $endDate);
                $totalAppropriationsDailyFromDatesAndHours = array();
                foreach($datesAndHours as $date => $hour) {
                    if(strpos($totalAppropriationsDaily, $date) === false){
                        array_push($totalAppropriationsDailyFromDatesAndHours, $date.',0');
                    }
                }

                $totalAppropriationsDailyCondition = is_array($totalAppropriationsDaily) ? count($totalAppropriationsDaily) : 0;

                if($totalAppropriationsDailyCondition){
                    $totalAppropriationsDaily .= ';'.implode(';', $totalAppropriationsDailyFromDatesAndHours);
                }


                $totalAppropriationsDaily = explode(';', $totalAppropriationsDaily);

                /** @var Repositories\WorkedTime $repositoryWorkTime */
                $repositoryWorkTime = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\WorkedTime');
                $totalDiff = 0;
                $totalWorkedTime = 0;
                foreach ($totalAppropriationsDaily as $totalAppropriationDaily) {
                    if($totalAppropriationDaily){
                        list($day, $totalAppropriation) = explode(',', $totalAppropriationDaily);

                        $dayHours = $datesAndHours[$day];

                        $workedTime = $repositoryWorkTime->getWorkTimeByUserAndDate($appropriation['ID'], $day);
                        if (empty($workedTime)) {
                            $diff = $dayHours - intval($totalAppropriation);
                            $totalWorkedTime += $dayHours;
                        } else {
                            $diff = intval($workedTime[0]['WORK_TIME']) - intval($totalAppropriation);
                            $totalWorkedTime += intval($workedTime[0]['WORK_TIME']);
                        }
                        if ($diff > 0) {
                            $appropriation['HOURS_CONTROL_NAME'] = -1;
                        } elseif ($diff < 0) {
                            if (empty($appropriation['HOURS_CONTROL_NAME'])) {
                                $appropriation['HOURS_CONTROL_NAME'] = 1;
                            }
                        }
                        $totalDiff += $diff;
                    }

                }
                if ($totalDiff < 0) {
                    $diffFront = '+';
                } elseif ($totalDiff == 0) {
                    $diffFront = '';
                } else {
                    $diffFront = '-';
                }
                if ($totalWorkedTime == 0) {
                    $appropriation['DIFF_APPROPRIATION'] = 0;
                } else {
                    $appropriation['DIFF_APPROPRIATION'] = $diffFront . round(abs(($totalDiff) * 100) / $totalWorkedTime, 2) . '%';
                }
                $appropriation['HOURS_CONTROL'] = $totalDiff;

                $structure = $this->structureService->getStructureByResource($resource);
                $appropriation['PARENT'] = empty($structure) ? null : $structure->getId();

                // Avoid ID duplicates on tree data
                if(!empty($appropriation['PARENT']))
                    $appropriation['PARENT'] = 'STRUCT'.$appropriation['PARENT'];

                $diffAppropriationsMeanByStructure = $this->calculateDiffAppropriationsMeanByStructureOnDemand($appropriation, $appropriation['PARENT'], $structuresIdsParentsIdsTree, $diffAppropriationsMeanByStructure);
            }

            $structures = $this->buildStructuresAppropriationsRows($structures, $structuresIds, $diffAppropriationsMeanByStructure);
            // Maybe is not necessary, because it without it prints the "real" situation of appropriation
            // including the structures without any appropriation
            //$structures = $this->filterStructuresWithAppropriationChildren($structures, $appropriations);


            $appropriations = array_merge($structures, $appropriations);
        }
        return $appropriations;
    }

    public function removeAppropriation($appropriation)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $appropriation = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Appropriation', $appropriation);
            $this->entityManager->remove($appropriation);
            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    public function getAppropriationSpendTimeByTask(\Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Calc\TaskInterface $task)
    {
        $query = <<<SQL
        SELECT SUM(AP.TOTAL) as TOTAL
                FROM APPROPRIATION AP 
                JOIN TASK TA ON (TA.ID = AP.TASK)
                WHERE TA.TREE_KEY like  :KEY
SQL;
        $params = array(
            'KEY' => $task->getTreeKey() . '%'
        );
        $connection = $this->entityManager->getConnection();
        $result = $connection->fetchAll($query, $params);
        $result[0]["TOTAL"] ?: 0;
        return $result[0]["TOTAL"];

    }

    public function findAppropriationByTask($idTask, $date = null)
    {
         return $this->appropriationRepository->findAppropriationByTask($idTask, $this->environment->getCurrentUserId(),
            $this->environment->getCurrentOrganizationId(), $date);
    }


    public function validateApropriationByTask($idTask, $date = null)
    {
         return $this->appropriationRepository->validateApropriationByTask($idTask);
    }

    public function getTaskAppropriationsTimeControl($idTask, $filters)
    {
        /** @var Repositories\Appropriation $appropriationRepository */
        $appropriationRepository = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Appropriation');
        $appropriations = $appropriationRepository->getAppropriationsByTaskTimeControl($idTask, $this->environment->getCurrentUserId(),
            $this->environment->getCurrentOrganizationId(), $filters);
        foreach ($appropriations as &$appropriation) {
            if ($appropriation['END_HOUR'] || $appropriation['BEGIN_HOUR']) {
                $dayWeek = intval(substr($appropriation['BEGIN_HOUR'], 0, 1));
                $dayWeek = \Service\Calendar\TaskCalendar::$ARRAY_WEEK_BR_LABEL[$dayWeek - 1];
                $appropriation['BEGIN_HOUR'] = $dayWeek . ' ' . substr($appropriation['BEGIN_HOUR'], 2);
                $appropriation['END_HOUR'] = $dayWeek . ' ' . substr($appropriation['END_HOUR'], 2);
            }
        }
        return $appropriations;
    }

    public function saveTaskPercentCompleted($appropriation)
    {
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            $idAllocation = $appropriation['ALLOCATION'];
            $percentCompleted = empty($appropriation['PERCENT_COMPLETED']) ? 1 : (int)$appropriation['PERCENT_COMPLETED'];
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation $allocation */
            $allocation = $this->allocationRepository->find($idAllocation);
            $task = $allocation->getTask();
            $task->setPercentCompleted($percentCompleted);
            $this->entityManager->persist($task);
            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $task;
    }

    public function findPaths($taskId, $resourceId)
    {
        $task = $this->taskRepository->find($taskId);
        $projects = $this->taskRelationshipRepository->findRelatedProjectIds($taskId);
        if(!empty($projects)) {
            $projectIds = $projects['RELATED_PROJECTS'] . ',' . $projects['MAIN_PROJECT'];
        } else {
            $projectIds = $task->getProject()->getId();
        }

        return $this->resourceRepository->getPaths($projectIds, $resourceId);
    }

    public function validateProjectRepositories($taskId)
    {
        $validation = array('IS_RELATED_TASK' => false);
        $task = $this->taskRepository->find($taskId);
        $projects = $this->taskRelationshipRepository->findRelatedProjectIds($taskId);
        if(!empty($projects)) {
            $projectIds = $projects['RELATED_PROJECTS'] . ',' . $projects['MAIN_PROJECT'];
            $validation['IS_RELATED_TASK'] = true;
            $relatedTaskId = $this->taskRelationshipRepository->findRelatedTaskIds($taskId);
            $validation['MAIN_TASK'] = $relatedTaskId['MAIN_TASK'];
            $mainTask = $this->taskRepository->find($relatedTaskId['MAIN_TASK']);
            $validation['MAIN_PROJECT'] = $mainTask->getProject()->getId();
        } else {
            $projectIds = $task->getProject()->getId();
        }
        $userId = $this->environment->getCurrentUserId();
        $validation['PROJECTS'] = $this->resourceRepository->validateProjectDirectories($projectIds,$userId);
        return $validation;
    }

    public function isFirstAppropriation($task, $resource)
    {
        $result = $this->appropriationRepository->findBy(array('task' => $task, 'resource' => $resource));
        return count($result) === 0 ? true : false;
    }

    public function deleteAppropriationById($idAppropriation)
    {
        $appropriation = $this->appropriationRepository->find($idAppropriation);
        $this->entityManager->remove($appropriation);
        $this->entityManager->flush();
    }

    public function getBeginHourAppropriation($taskId)
    {
        $userId = $this->environment->getCurrentUserId();
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation $allocation */
        $allocation = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation')
            ->findOneBy(array('task' => $taskId, 'resource' => $userId));
        $playDate = '';
        if (!empty($allocation) && $allocation->getPlayDate()) {
            if (Utilities::isDateToday($allocation->getPlayDate())) {
                $playDate = $allocation->getPlayDate()->format('d/m/Y H:i:s');

            }
        }
        return array('PLAY_DATE' => $playDate);
    }

    private function applyOrderToIssues(Entities\Task $task, $insertTaskInIssue = false)
    {
        /** @var Entities\TaskRelationship[] $taskRelationships */
        $taskRelationships = $task->getTaskRelationshipInverse();
        if (!empty($taskRelationships)) {
            $this->setStatusIssues($insertTaskInIssue, $taskRelationships);
        }
    }


    /**
     * @param $idAppropriation
     * @param $idAllocation
     * @param $beginHour
     * @param $endHour
     * @param $total
     * @param $description
     * @param $percentCompleted
     * @param $specificationComment
     * @param $specificationCommentType
     * @param $specificationDocument
     * @param $specificationCdItemtrab
     * @return Entities\Appropriation
     * @throws \Exception
     */
    public function saveAppropriation($idAppropriation, $idAllocation, $beginHour, $endHour, $total, $description,
                                      $percentCompleted,$specificationComment,$specificationCommentType,
                                      $specificationDocument, $specificationCdItemtrab)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();

            if(empty($idAllocation)){
                throw new \Exception(Messages::ERR_RESOURCE_NOT_ALLOCATED_IN_TASK);
            }

            /** @var Entities\Allocation $allocation */
            $allocation = $this->allocationRepository->find($idAllocation);
            $resource   = $allocation->getResource();
            $task       = $allocation->getTask();
            $project    = $task->getProject();
            /** @var Entities\Task[] $relatedTasks */
            $relatedTasks = array();
            $this->validateRelatedTaskSaveAppropriation($task, $relatedTasks);

            if($project->getProjectType() && $project->getProjectType()->getAllowendtask() == ProjectType::IS_NOT_ALLOW_END_TASK){
                $percentCompleted = 1;
            }

            /** @var Entities\Appropriation $appropriation */
            $appropriation = $this->createOrUpdateAppropriationSaveAppropriation($idAppropriation, $beginHour, $endHour, $total,
                                                                $description, $percentCompleted, $task, $resource, $relatedTasks);

            if ($appropriation->isOneHundredPercent() && !$task->getEndDate()) {
                $task->setEndDate($beginHour);
                foreach($relatedTasks as &$relatedTask) {
                    $relatedTask->setEndDate($beginHour);
                }
            }

            if ($task->getProject()->getProjectType() && $task->getProject()->getProjectType()->getAllowendtask() == ProjectType::IS_NOT_ALLOW_END_TASK) {
                $percentCompleted = 1;
            } else {
                $percentCompleted = $this->appropriationRepository->getMaxPercentCompleted($task->getId(), $resource->getId());
                if ($percentCompleted < $appropriation->getCompleted()) {
                    $percentCompleted = $appropriation->getCompleted();
                }
            }
            if ($task->getTaskStatus()->getControl() == Task::TASK_STATUS_CONTROL_CONCLUDED) {
                $percentCompleted = 100;
            }
            $task->setPercentCompleted($percentCompleted);
            $latestDate = $this->appropriationRepository->getLatestAppropriationDate($task->getId(),$idAppropriation);
            $latestDate = \DateTime::createFromFormat(Utilities::FORMAT_DATE_DEFAULT, $latestDate[0]["END_HOUR"]);
            if ($task->getPercentCompleted() == 100 && (($endHour && $endHour > $latestDate) || (!$task->getEndDate()))) {
                $task->setEndDate($endHour);
                foreach($relatedTasks as &$relatedTask) {
                    $relatedTask->setEndDate($endHour);
                }
            }
            foreach($relatedTasks as &$relatedTask) {
                $relatedTask->setPercentCompleted($percentCompleted);
            }

            $this->setTaskStatusSaveAppropriation($percentCompleted, $allocation, $task, $relatedTasks, $resource);

            $this->entityManager->persist($allocation);
            $this->entityManager->persist($appropriation);
            $this->entityManager->persist($task);
            foreach($relatedTasks as &$relatedTask) {
                $this->entityManager->persist($relatedTask);
            }

            $this->entityManager->flush();


            $this->commentInSpecification($task,$specificationComment,$specificationCommentType,$specificationDocument, $resource->getExternalCode(), $specificationCdItemtrab);


            $this->applyOrderToIssues($task);
            foreach($relatedTasks as &$relatedTask) {
                $this->applyOrderToIssues($relatedTask);
            }

            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $appropriation;
    }

    public function  createdTaskClientAccept(Entities\Task $issue, Entities\Project $project){
        Utilities::getTaskService()->createdTaskClientAccept($issue,$project);
    }

    public function integrationSMFinishSpecification($issueId, $cdOperador = null) {
        /** @var Entities\Budget $budget */
        $budget = InstanceProvider::getRepository()->getBudget()->findOneBy(array('task' => $issueId));
        $organizationId = $this->environment->getCurrentOrganizationId();
        $budgetRepository = InstanceProvider::getRepository()->getBudget();
        $urlSm = $budgetRepository->getUrlIntegrationSM($organizationId);
        if ($urlSm && $budget) {
            if ($budget->getExternalId()) {
                Integration\IntegrationSM::integrationSMCancel($issueId, $organizationId);
            }
            $proposalId = Integration\IntegrationSM::integrationSMFinishSpecification($issueId, $organizationId);
            $nrparcnegocio = $this->taskRelationshipRepository->findNrparcnegocioByCdoperator($cdOperador);
            if (!empty($nrparcnegocio) && $proposalId) {
                $nrparcnegocio = $nrparcnegocio[0]['NRPARCNEGOCIO'];
                $cdoperador = $this->environment->getCurrentCdoperador();
                $vendedorProposta = $this->fctVendedororcaFactory->save(null, $proposalId, $nrparcnegocio,
                    $organizationId, $cdoperador, $cdoperador);
                $this->entityManager->persist($vendedorProposta);
                $this->entityManager->flush();
            }
            $budget->setExternalId($proposalId);
            $this->entityManager->persist($budget);
            $this->entityManager->flush();

        }
    }

    public function commentInSpecification(Entities\Task $task,$specificationComment,$specificationCommentType,
                                           $specificationDocument, $cdOperador = null, $specificationCdItemtrab = []){
        if($task->getClassification()->getControl() == Entities\Classification::CONTROL_SPECIFICATION
         && $specificationComment ){
            /** @var Entities\TaskRelationship[] $issueRelationships */
            $issueRelationships = $this->taskRelationshipRepository->findBy(array('taskRelationship' => $task,'type' => Entities\TaskRelationship::RELATIONSHIP_ISSUE_TYPE));
            foreach ($issueRelationships as $issueRelationship) {
                Utilities::getCommentService()->commentInSpecification($issueRelationship->getTask(),
                    $specificationComment,$specificationCommentType,$specificationDocument);
                InstanceProvider::getService()->getTaskWorkitem()->save($issueRelationship->getTask()->getId(), $specificationCdItemtrab);

                $this->createdTaskClientAccept($issueRelationship->getTask(),$issueRelationship->getTask()->getProject());

                $taskClient = $this->taskRelationshipRepository->findTaskClient($task->getId());
                if($taskClient && $taskClient["EXTERNAL_CODE"] !== $this->cdClienteTek || !$taskClient){
                    $this->integrationSMFinishSpecification($issueRelationship->getTask()->getId(), $cdOperador);
                }

            }
        }

    }

    /**
     * @param $taskId
     * @param $currentTaskId
     * @param $forceCreateBranch
     * @param $originProductVersionId
     * @throws \Doctrine\DBAL\ConnectionException
     * @throws \Exception
     */
    public function insertDevOpsBranch($taskId, $currentTaskId, $forceCreateBranch = null, $originProductVersionId = null)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $this->environment->setDOCurrentUserIdByCurrentPMUser();
            $relatedTasks = $this->taskRelationshipRepository->findRelatedTaskIds($taskId);
            if(!empty($relatedTasks)) {
                $taskId = $relatedTasks['MAIN_TASK'];
                $relatedProjectIds = $this->taskRelationshipRepository->findRelatedProjectIdsWithResourceTaskFilter($taskId, $currentTaskId);
                if(empty($relatedProjectIds['MAIN_PROJECT'])) {
                    $projectIdsStr = $relatedProjectIds['RELATED_PROJECTS'];
                } else if(empty($relatedProjectIds['RELATED_PROJECTS'])) {
                    $projectIdsStr = $relatedProjectIds['MAIN_PROJECT'];
                } else {
                    $projectIdsStr = $relatedProjectIds['MAIN_PROJECT'] . ',' . $relatedProjectIds['RELATED_PROJECTS'];
                }

                $projectIds = explode(',', $projectIdsStr);
            } else {
                $task = $this->taskRepository->find($taskId);
                $projectIds = array($task->getProject()->getId());
            }
            /** @var Entities\TaskBranches $taskBranch */
            $taskBranch = $this->taskBranchesRepository->findOneBy(array('task' => $taskId));
            $branchName = $taskBranch->getName();
            if(strpos($branchName, 'TASK_BKP') === false || $forceCreateBranch == true) {
                $repositories = [];
                /** @var Entities\Task $task */
                $task = $this->taskRepository->find($taskId);
                $productVersionIdDev = $task->getProductVersionDev() ? $task->getProductVersionDev()->getId() : null;
                $repositories = $this->getChangedFilesRepositoriesByTask($taskId, $repositories, $projectIds);
                $subreleases = $this->repositoryRepository->findSubReleasesAndRepositoriesByRepositoryIds($repositories, $productVersionIdDev, $originProductVersionId);

                foreach($subreleases as $subrelease) {
                    $subreleaseId = $subrelease['SUBRELEASE_ID'];
                    $repositoriesBranch = explode(',', $subrelease['REPOSITORY_IDS']);
                    /** @var ConfigurationEntities\Subrelease $subrelease */
                    $subrelease = $this->subreleaseRepository->find($subreleaseId);
                    $product = $subrelease->getRelease()->getProduct();
                    $subrelease->setStatus(self::SUBRELEASE_STATUS_PENDING);
                    $this->entityManager->persist($subrelease);
                    $this->createBranchAndRepository($branchName, $subrelease, $product, $repositoriesBranch);
                }
            }
            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {

            $connection->rollBack();
            throw $e;
        }
    }

    /**
     * @param Entities\Repository[] $repositories
     * @param Entities\Directory[] $directories
     * @return array
     */
    public function filterRepositoriesByDirectories(array $repositories,array $directories){
        $returnRepositories = array();
        foreach ($directories as $directory) {
            if (in_array($directory->getRepository()->getId(), $repositories)) {
                $returnRepositories[] = $directory->getRepository()->getId();
            }
        }
        return $returnRepositories;
    }

    private function validateCheckSameDirectoryFromDifferentRepositories($arrayPathVerification, $arrayPathVerificationAux, $isWindows){
        if($isWindows){
            if(($arrayPathVerification["LOCAL_DIRECTORY"] == $arrayPathVerificationAux["LOCAL_DIRECTORY"]) &&
               ($arrayPathVerification["repository"] != $arrayPathVerificationAux["repository"] ) ){
                return true;
            }
            return false;
        }else{
            if(($arrayPathVerification["LINUX_DIRECTORY"] == $arrayPathVerificationAux["LINUX_DIRECTORY"]) &&
               ($arrayPathVerification["repository"] != $arrayPathVerificationAux["repository"])){
                return true;
            }
            return false;
        }
    }

    public function checkSameDirectoryFromDifferentRepositories($taskId, $typeRelationship, $isWindows)
    {
        $task = $this->taskRepository->find($taskId);
        $resourceId = $this->environment->getCurrentUserId();
        $projects = $this->getRestrictProjectIdsForTaskMergeRequest($taskId, $typeRelationship);
        $arrayPaths = $this->findPaths($task->getId(), $resourceId, $projects);
        $repositoryURL = array();
        $projectIdMessage = array();
        $projectDiretories = array();
        $arrayPathVerifications = $arrayPaths;
        foreach ($arrayPathVerifications as $arrayPathVerification) {
            foreach ($arrayPathVerifications as $arrayPathVerificationAux) {
                if($this->validateCheckSameDirectoryFromDifferentRepositories($arrayPathVerification, $arrayPathVerificationAux, $isWindows)){
                    if(!in_array($arrayPathVerification['ID'][0], $projectIdMessage)){
                        $projectIdMessage[] = $arrayPathVerification['ID'][0];
                    }
                    if($isWindows){
                        if(!in_array($arrayPathVerification['LOCAL_DIRECTORY'], $projectDiretories)){
                            $projectDiretories[] = $arrayPathVerification['LOCAL_DIRECTORY'];
                        }
                    }else{
                        if(!in_array($arrayPathVerification['LINUX_DIRECTORY'], $projectDiretories)){
                            $projectDiretories[] = $arrayPathVerification['LINUX_DIRECTORY'];
                        }
                    }
                }
            }
        }
        $message = '';
        $projects = array();
        if(count($projectIdMessage) > 0){
            $message = \Util\Messages::ERR_SAME_DIRECTORY_IN_DIFERENT_REPOSITORIES;
            $projectDiretories = array_unique($projectDiretories);
            foreach ($projectDiretories as $projectDiretorie) {
                $message = $message.'<br/><br/><li>'.$projectDiretorie.'</li>';
                foreach ($arrayPaths as $arrayPath) {
                    if(in_array($arrayPath['ID'][0], $projectIdMessage) && ($arrayPath['LOCAL_DIRECTORY'] == $projectDiretorie || $arrayPath['LINUX_DIRECTORY'] == $projectDiretorie)){
                        $project = InstanceProvider::getRepository()->getProject()->find($arrayPath['PRODUCT_ID']);
                        $message = $message.'<br/>Projeto: "'.$project->getName().'"<br/>Repositório: "'.$arrayPath['NAME'].'"<br/>';
                    }
                }
            }
            throw new \Exception($message);
        }else{
            return $projects;
        }
    }

    public function checkAllChangedFilesHaveLocalConnection($taskId,$typeRelationship)
    {
        /** @var Entities\Task $task */
        $task = $this->taskRepository->find($taskId);
        $relatedProjectIds = $this->taskRelationshipRepository->findRelatedProjectIds($taskId);
        $resourceId = $this->environment->getCurrentUserId();
        $projects = $this->getRestrictProjectIdsForTaskMergeRequest($taskId, $typeRelationship);
        $arrayPaths = $this->findPaths($task->getId(), $resourceId, $projects);
        $repositoryURL = array();
        foreach ($arrayPaths as $arrayPath) {
                $repositoryURL[] = $arrayPath['repository'];
        }
        $changedFiles = $this->changedFilesRepository->findBy(array('task' => $task));
        $repositories = $this->getNameRepositoriesWithoutChangedFiles($changedFiles, $repositoryURL,$projects,$relatedProjectIds );
        return $repositories;
    }

    public function findTaskTypesForFraming(Entities\Task $issue){
        return Utilities::getTaskService()->findTaskTypesForFraming($issue);
    }

    public function findProjectTypesForFraming(Entities\Task $issue){
        return InstanceProvider::getRepository()->getTask()->findProjectTypesForFraming($issue->getId());
    }
    public function validateFraming(Entities\Task $issue){
        return $issue->getTaskStatus() &&
               $issue->getTaskStatus()->getTaskType() &&
               !in_array($issue->getTaskStatus()->getTaskType()->getControl(),
                         array(Entities\TaskType::CONTROL_HOMOLOGATED, Entities\TaskType::CONTROL_BACKLOG));
    }
    public function validateIssueHomologated(Entities\Task $issue){
        return $issue->getTaskStatus() &&
               $issue->getTaskStatus()->getTaskType() &&
               $issue->getTaskStatus()->getTaskType()->getControl() == Entities\TaskType::CONTROL_HOMOLOGATED;
    }

    /**
     * @param $insertTaskInIssue
     * @param Entities\TaskRelationship[] $taskRelationships
     * @param bool $isFlush
     * @param bool $delete
     */
    public function setStatusIssues($insertTaskInIssue, $taskRelationships,$isFlush = true, $delete = false)
    {
        $organizationId  = $this->environment->getCurrentOrganizationId();
        foreach ($taskRelationships as $taskRelationship) {
            if ($taskRelationship->getType() == TaskPrecursor::RELATIONSHIP_ISSUE_TYPE) {
                $issue = $taskRelationship->getTask();
                $taskTypeForFraming =  $this->findTaskTypesForFraming($issue);
                $projectTypeForFraming =  $this->findProjectTypesForFraming($issue);
                $notHomologated = !$this->validateIssueHomologated($issue);
                if($taskTypeForFraming && $notHomologated){
                    $this->setStatusFraming($issue, $taskTypeForFraming);
                    $issue->setCurrentTaskStatusIn((new \DateTime()));
                }else if($projectTypeForFraming && $notHomologated) {

                    $this->setStatusFraming($issue, $projectTypeForFraming);
                }else if($insertTaskInIssue || $this->validateFraming($issue)) {
                    $frameOrder = $this->taskRepository->getIssueFrameOrderBasedOnRelatedTasks($issue->getId());
                    if ($frameOrder != null) {
                        /** @var Entities\TaskType $taskType */
                        $taskType = $this->taskTypeRepository->findOneBy(array('orderToFrame' => $frameOrder,
                            'organization' => $this->environment->getCurrentOrganizationId()));
                        if($delete){
                            /** @var Entities\TaskStatus $newTaskStatus */
                            $newTaskStatus = $this->taskStatusRepository->findOneBy(array('control' => self::CONTROL_BACKLOG,
                                'usedInIssue' => self::USED_IN_ISSUE_YES,
                                'organization' => $organizationId));
                        }else {
                            /** @var Entities\TaskStatus $newTaskStatus */
                            $newTaskStatus = $this->taskStatusRepository->findOneBy(array('taskType' => $taskType->getId(),
                                'usedInIssue' => self::USED_IN_ISSUE_YES,
                                'organization' => $organizationId));
                        }
                        if($issue->getTaskStatus() != $newTaskStatus){
                        	$issue->setCurrentTaskStatusIn((new \DateTime()));
                        }

                        $issue->setTaskStatus($newTaskStatus);
                        $this->entityManager->persist($issue);
                        if ($isFlush) {
                            $this->entityManager->flush();
                        }
                    }
;
                }

            }
        }
    }

    /**
     * @param $fileName
     * @param $fileContents
     * @throws \Exception
     */
    public function createLogFile($fileName, $fileContents)
    {
        $file = $this->logDirFile . DIRECTORY_SEPARATOR . $fileName;
        try {
            file_put_contents($file, $fileContents, FILE_APPEND);
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * @param $fileName
     * @throws \Exception
     * @return string
     */
    public function getLogFileContents($fileName)
    {
        $file = $this->logDirFile . DIRECTORY_SEPARATOR . $fileName;
        try {
            return file_exists($file) ? [file_get_contents($file)] : [];
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * @param $task
     * @param $relatedTasks
     */
    private function validateRelatedTaskSaveAppropriation(&$task, &$relatedTasks)
    {
        /** @var Entities\Task $task */
        $relatedTaskIds = $this->taskRelationshipRepository->findRelatedTaskIds($task->getId());
        if (!empty($relatedTaskIds)) {
            $relTaskIds = explode(',', $relatedTaskIds['RELATED_TASKS'] . ',' . $relatedTaskIds['MAIN_TASK']);
            foreach ($relTaskIds as $relTaskId) {
                $relatedTasks[] = $this->taskRepository->find($relTaskId);
            }
        }
    }

    /**
     * @param $percentCompleted
     * @param Entities\Allocation $allocation
     * @param Entities\Task $task
     * @param Entities\Task[] $relatedTasks
     * @param $resource
     * @throws \Exception
     */
    private function setTaskStatusSaveAppropriation($percentCompleted, $allocation, &$task, &$relatedTasks, $resource)
    {
        if ($percentCompleted == 100) {
            /* @var $taskStatusConcluded Entities\TaskStatus */
            $taskStatusConcluded = $this->taskStatusRepository->findOneBy(array('control' => Task::TASK_STATUS_CONTROL_CONCLUDED, 'isTask' => 'Y'));
            if (empty($taskStatusConcluded)) throw new \Exception(\Util\Messages::ERR_NOT_STATUS_TASK_BY_CONTROL);
            $allocation->setTaskStatus($taskStatusConcluded);
            $task->setTaskStatus($taskStatusConcluded);

            foreach ($relatedTasks as &$relatedTask) {
                $relatedTask->setTaskStatus($taskStatusConcluded);
                $relatedAllocation = $this->allocationRepository->findOneBy(array('resource' => $resource, 'task' => $relatedTask));
                $relatedAllocation->setTaskStatus($taskStatusConcluded);
                $this->entityManager->persist($relatedAllocation);
            }
        } else if ($task->getTaskStatus()->getControl() === Task::TASK_STATUS_CONTROL_NOT_INITIALIZED) {
            /* @var $taskStatus Entities\TaskStatus */
            $taskStatus = $this->taskStatusRepository->findOneBy(array('control' => Task::TASK_STATUS_CONTROL_PAUSED, 'isTask' => 'Y'));
            if (empty($taskStatus)) throw new \Exception(\Util\Messages::ERR_NOT_STATUS_TASK_BY_CONTROL);
            $allocation->setTaskStatus($taskStatus);
            $task->setTaskStatus($taskStatus);

            foreach ($relatedTasks as &$relatedTask) {
                $relatedTask->setTaskStatus($taskStatus);
                $relatedAllocation = $this->allocationRepository->findOneBy(array('resource' => $resource, 'task' => $relatedTask));
                $relatedAllocation->setTaskStatus($taskStatus);
                $this->entityManager->persist($relatedAllocation);
            }
        }
    }

    /**
     * @param $idAppropriation
     * @param $beginHour
     * @param $endHour
     * @param $total
     * @param $description
     * @param $percentCompleted
     * @param Entities\Task $task
     * @param $resource
     * @param Entities\Task[] $relatedTasks
     * @return Entities\Appropriation
     */
    private function createOrUpdateAppropriationSaveAppropriation($idAppropriation, $beginHour, $endHour, $total, $description,
                                                                  $percentCompleted, &$task, $resource, &$relatedTasks)
    {
        if (empty($idAppropriation)) {
            $appropriation = $this->appropriationFactory->createAppropriation($this->environment->getCurrentOrganization(),
                $task, $resource, $total, $beginHour, $endHour, $percentCompleted, $description);

            if ($this->isFirstAppropriation($task, $resource)) {
                $task->setBeginDate($appropriation->getBeginHour());
                foreach ($relatedTasks as &$relatedTask) {
                    $relatedTask->setBeginDate($appropriation->getBeginHour());
                }
            }

            if ($percentCompleted === self::MAX_APPROPRIATION_PERCENT) {
                $task->setResponse($description);
                foreach ($relatedTasks as &$relatedTask) {
                    $relatedTask->setResponse($description);
                }
            }
        } else {
            $appropriationRepository = $this->appropriationRepository;
            /** @var Entities\Appropriation $appropriationEntity */
            $appropriationEntity = $appropriationRepository->find($idAppropriation);
            $currentCompleted = $appropriationEntity->getCompleted();

            if ($appropriationEntity->isOneHundredPercent()) {
                $percentCompleted = 100;
            }

            if ($currentCompleted < $percentCompleted && $percentCompleted == self::MAX_APPROPRIATION_PERCENT) {
                $task->setResponse($description);
                foreach ($relatedTasks as &$relatedTask) {
                    $relatedTask->setResponse($description);
                }
            }

            $appropriation = $this->appropriationFactory->updateAppropriation($appropriationEntity, $total, $beginHour,
                $endHour, $percentCompleted, $description);
        }

        return $appropriation;
    }

    /**
     * @param $branchName
     * @param $subrelease
     * @param $product
     * @param $repositories
     * @throws \Doctrine\ORM\OptimisticLockException
     */
    private function createBranchAndRepository($branchName, $subrelease, $product, $repositories)
    {
        $branch = $this->branchRepository->findOneBy(array('name' => $branchName, 'subrelease' => $subrelease));
        if (empty($branch)) {
            $branch = $this->branchFactory->saveBranch(null, $product, $subrelease, $branchName, Messages::DEO_CREATE_DESCRIPTION_DEFAULT);
            $this->entityManager->persist($branch);
            $this->entityManager->flush();
        }

        foreach ($repositories as $repositoryId) {
            /** @var Entities\Repository $repository */
            $repository = $this->repositoryRepository->find($repositoryId);
            $repositoryBranch = $this->repositoryBranchFactory->saveRepositoryBranch(null, $branch, $repository);
            $this->entityManager->persist($repositoryBranch);
        }
    }

    /**
     * @param $taskId
     * @param $repositories
     * @param $projectIds
     * @return array
     */
    private function getChangedFilesRepositoriesByTask($taskId, $repositories, $projectIds)
    {
        $changedFiles = $this->changedFilesRepository->findBy(array('task' => $taskId));

        /** @var Entities\ChangedFiles $changedFile */
        foreach ($changedFiles as $changedFile) {
            $repository = $changedFile->getRepository();
            if (!in_array($repository->getId(), $repositories) &&
                in_array($repository->getProject()->getId(), $projectIds)) {
                $repositories[] = $repository->getId();
            }
        }
        return $repositories;
    }

    public function findDateLastAppropriation()
    {
        $dateLastAppropriation = $this->appropriationRepository->findDateLastAppropriation($this->environment->getCurrentUserId());
        $dateLastAppropriation = (isset($dateLastAppropriation)) ? $dateLastAppropriation : date('d/m/Y H:i');
        return $dateLastAppropriation;
    }

}
