var NAMESPACE = 'pm_do';
function GitQueue(ZhSocketFactory, ZHPromise, ScreenService) {
    this.currentRoom = "";
    var GitResponse = function() {
        this.progress = function(data) {
            var innerPromise = currentPromises[data.id];
            if(innerPromise) {
                innerPromise.defer.notify(data.data);
            }
        };
        this.noOtherClientsConnected = function(data) {
            try {
                clearTimeout(timeOut);
                var innerPromise = currentPromises[data.id];
                responses[data.id] = true;
                log(innerPromise);
                ScreenService.showMessage("Project Services desconectado! Conecte-o e tente novamente.");
                data.connectionError = true;
                rejectConnectionError(innerPromise, data);
            } finally {
                delete currentPromises[data.id];
            }
        };
        var rejectConnectionError = function(innerPromise, data) {
            if (innerPromise) {
                if (!innerPromise.ignoreError) {
                    innerPromise.defer.reject(data);
                } else {
                    innerPromise.defer.resolve({});
                }
            }
        };
        this.response = function(data) {
            clearTimeout(timeOut);
            var innerPromise = currentPromises[data.id];
            responses[data.id] = true;
            log(innerPromise);
            try {
                if (data.isError) {
                    if (data.connectionError) {
                        ScreenService.showMessage("Project Services desconectado! Conecte-o e tente novamente.");
                    } else if(data.action == "shExec"){
                        return innerPromise.defer.reject(data);
                    }
                    rejectConnectionError(innerPromise, data);
                } else {
                    if (innerPromise) {
                        if (innerPromise.rejectIfError && data.errors > 0) {
                            innerPromise.defer.reject(data);
                        }else{
                            innerPromise.defer.resolve(data.data);
                        }
                    }
                }
            } finally {
                delete currentPromises[data.id];
            }
        };
    };
    var self = this;
    var currentId = 0;
    var defaultTimeout = 30000;
    var timeOut;
    var LOGIN_BROWSER = "loginBrowser";
    var queueRuning = false;
    var responses = {};
    var host = 'ws://localhost:8080/';
    var webSocket;
    var currentPromises = {};
    var currentDirectory;
    this.pendingPayload = [];
    var sendPayloadMutex = false;
    var _sendPayloadFinishedFactory = function(payload) {
        return (function() {
            if (responses[payload.data.id]) {
                if (!this.__runed) {
                    this.__runed = true;
                    sendPayloadMutex = false;
                }
            }
        }).bind({});
    };
    this.hasQueueMessage = function() {
        return this.pendingPayload.length > 0;
    };

    this._callAsyncQueue = function() {
        setTimeout((function() {
            this._callQueue();
        }).bind(this), 100);
    };
    this._callQueue = function() {
        if (this.hasQueueMessage()) {
            if (!sendPayloadMutex) {
                var currentPayload = this.pendingPayload.shift();
                this._execSendPayload(currentPayload);
            }
            if (this.hasQueueMessage())
                this._callAsyncQueue();
        }
    };

    this._execSendPayload = function(payload) {
        if (!sendPayloadMutex) {
            if(payload.useMutex) {
                sendPayloadMutex = true;
            }
            var dtoPayload = payload;
            dtoPayload.room = this.currentRoom;
            dtoPayload.namespace = NAMESPACE;
            var defer = payload.defer;
            var resolve = _sendPayloadFinishedFactory(payload);
            if(payload.data && payload.data.timeout) {
                timeOut = setTimeout(function() {
                    defer.reject({
                        connectionError: true
                    });
                    ScreenService.hideLoader();
                    ScreenService.restoreDefaultLoadingMessage();
                    ScreenService.showMessage("O comando executado no WebSocket demorou muito para responder. <br> Verifique o log do serviço e da aplicação, atualize a página do seu navegador e tente novamente.");
                }, payload.data.timeout);
            }
            webSocket.send(JSON.stringify(dtoPayload));
            defer.promise.then(resolve, resolve);
        }
    };

    this._sendPayload = function(payload) {
        this.pendingPayload.push(payload);
        this._callAsyncQueue();
    };

    this._factoryPayload = function(payload) {
        var defer = ZHPromise.defer();
        currentId++;
        payload.data.id = currentId;
        payload.defer = defer;
        payload.defer.promise.progress = (function(callback){
            payload.defer.promise.callbackProgress = callback;
            return this;
        }).bind(this);

        payload.defer.promise.notify = function(data){
            if(payload.defer.promise.callbackProgress){
                payload.defer.promise.callbackProgress(data);
            }
        };
        return payload;
    };

    this._decorate = function(payload) {
        this.then = payload.defer.promise.then.bind(payload.defer.promise);
        payload.defer.notify = function(data){
            if(payload.defer.promise.notify){
                payload.defer.promise.notify(data);
            }
        };
        this.progress = payload.defer.promise.progress.bind(payload.defer.promise);
        this.catch = payload.defer.promise.catch.bind(payload.defer.promise);
    };
    var _getLogMessage = function(data) {
        var message = '%cAction:%c ' + data.action + ' | %cResponse:%c ';

        if(typeof data.data == 'object') {
            message += '["' + data.data.join('", "') + '"] | ';
        } else {
            var value = data.data.replace(/\n/g, ' ');
            if(value !== '') {
                value = value.trim();
            }
            message += '"' + value + '" | ';
        }

        message += data.isError ? '%cError' : '%cSuccess';

        return message;
    };
    this.connect = function(host){
        webSocket = ZhSocketFactory.create(GitResponse, host);
        webSocket.addListener({
            'onMessage': function (data) {
                //Jump progress messages;
                if(data.data instanceof Array || typeof data.data == 'string') {
                    var errorSuccessRule = data.isError ? "color: red;" : "color: green;";
                    console.log(_getLogMessage(data), "color: blue;", "color: black;", "color: blue;", "color: black;", errorSuccessRule);
                }
            },
            'onOpen': function() {
                if(this.currentRoom) {
                    self.loginBrowser(this.currentRoom);
                }
            }
        });
    };

    this.sendPayload = function(prePayload) {
        var payload = this._factoryPayload(prePayload);
        currentPromises[payload.data.id] = payload;
        this._sendPayload(payload);
        this._decorate(payload);
        return this;
    };

    this.joinRoom = function(namespace, email) {
        var eventsToStart = ['createNamespace', 'createRoom', 'joinRoom'];
        var payload = {
            event: LOGIN_BROWSER,
            namespace: namespace,
            room: email,
            data: {
                'email': email,
                'key': 'you can create!'
            }
        };

        eventsToStart.forEach(function(event) {
            payload.event = event;
            webSocket.send(JSON.stringify(payload));
        });
        this.currentRoom = email;
        var room = { isRoom : true };
        for(var i in this) {
            if(Util.isFunction(this[i])) {
                room[i] = this[i].bind(room);
            } else {
                room[i] = this[i];
            }
        }
        delete room.catch;
        delete room.progress;
        delete room.then;
        room.pendingPayload = [];
        return room;
    };

    this.loginBrowser = function(email) {
        var defer = ZHPromise.defer();
        var timeout = setTimeout(function() {
            defer.reject({
                connectionError: true
            });
            ScreenService.hideLoader();
            ScreenService.restoreDefaultLoadingMessage();
            ScreenService.showMessage("A conexão com o WebSocket demorou muito para ser estabelecida. <br> Reinicie o seu serviço, atualize a página do seu navegador e tente novamente.");
        }, 30000);

        var room = this.joinRoom("pm_do", email);
        room.sendPayload({
            useMutex: true,
            event: "getPrivateRoom",
            data: {
                email: email
            }
        }).then(function(privateRoom) {
            clearTimeout(timeout);
            return self.joinRoom("pm_do", privateRoom);
        }).then(function(room) {
            defer.resolve(room);
        }).catch(function(error) {
            defer.reject(error);
        });

        return defer.promise;
    };

    this.add = function(currentDirectory, files, maxLength) {
        return this.sendPayload({
            event: "add",
            data: {
                dir: currentDirectory,
                files: files,
                maxLength: maxLength,
                timeout: defaultTimeout
            }
        });
    };
    this.commit = function(currentDirectory, message) {
        return this.sendPayload({
            event: "commit",
            data: {
                message: message,
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.commitWithAuthor = function(currentDirectory, message, authorName, authorEmail) {
        return this.sendPayload({
            event: "commitWithAuthor",
            data: {
                message: message,
                dir: currentDirectory,
                authorName: authorName,
                authorEmail: authorEmail,
                timeout: defaultTimeout
            }
        });
    };
    this.checkout = function(currentDirectory, branchName) {
        return this.sendPayload({
            event: "checkout",
            data: {
                branchName: branchName,
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.branch = function(currentDirectory) {
        return this.sendPayload({
            event: "branch",
            data: {
                dir: currentDirectory,
                timeout: 60000
            }
        });
    };
    this.branchSpecificRemote = function(currentDirectory, branchName) {
        return this.sendPayload({
            event: "branchSpecificRemote",
            data: {
                dir: currentDirectory,
                branchName: branchName,
                timeout: defaultTimeout
            }
        });
    };
    this.currentBranch = function(currentDirectory) {
        return this.sendPayload({
            ignoreError: true,
            event: "currentBranch",
            data: {
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };

    this.clone = function(currentDirectory, url) {
        return this.sendPayload({
            event: "clone",
            data: {
                url: url,
                dir: currentDirectory,
                timeout: 600000
            }
        });
    };
    this.removeLocalBranch = function(currentDirectory, name) {
        return this.sendPayload({
            event: "removeLocalBranch",
            data: {
                name: name,
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.removeRemoteBranch = function(currentDirectory, name) {
        return this.sendPayload({
            event: "removeRemoteBranch",
            data: {
                name: name,
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.kill = function(pid) {
        return this.sendPayload({
            event: "kill",
            //rejectIfError: true,
            data: {
                pid: pid
            }
        });
    };
    this.shExec = function(currentDirectory, file) {
        return this.sendPayload({
            event: "shExec",
            //rejectIfError: true,
            data: {
                file: file,
                dir: currentDirectory,
                timeout: 3600000
            }
        });
    };
    this.phpExec = function(currentDirectory, file) {
        return this.sendPayload({
            event: "phpExec",
            data: {
                file: file,
                dir: currentDirectory,
                timeout: 3600000
            }
        });
    };
    this.createBranch = function(currentDirectory, newBranchName) {
        return this.sendPayload({
            ignoreError: true,
            event: "createBranch",
            data: {
                newBranchName: newBranchName,
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.createPath = function(currentDirectory) {
        return this.sendPayload({
            event: "createPath",
            data: {
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.createTag = function(currentDirectory, nameTag) {
        return this.sendPayload({
            event: "createTag",
            data: {
                nameTag: nameTag,
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.merge = function(currentDirectory, nameBranch) {
        return this.sendPayload({
            event: "merge",
            data: {
                nameBranch: nameBranch,
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.pull = function(currentDirectory, branchName) {
        return this.sendPayload({
            event: "pull",
            data: {
                branchName: branchName,
                dir: currentDirectory,
                timeout: 60000
            }
        });
    };
    this.pullGeneric = function (currentDirectory) {
        return this.sendPayload({
            event: "pullGeneric",
            data: {
                dir: currentDirectory,
                timeout: 60000
            }
        });
    };
    this.pullNoCommit = function(currentDirectory, branchName) {
        return this.sendPayload({
            event: "pullNoCommit",
            data: {
                branchName: branchName,
                dir: currentDirectory,
                timeout: 60000
            }
        });
    };
    this.push = function(currentDirectory, branchName) {
        return this.sendPayload({
            event: "push",
            data: {
                branchName: branchName,
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.pushTag = function(currentDirectory) {
        return this.sendPayload({
            event: "pushTag",
            data: {
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.stash = function(currentDirectory) {
        return this.sendPayload({
            event: "stash",
            data: {
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.stashApply = function(currentDirectory) {
        return this.sendPayload({
            event: "stashApply",
            data: {
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.stashWithBranch = function(currentDirectory, branchName) {
        return this.sendPayload({
            event: "stashWithBranch",
            data: {
                dir: currentDirectory,
                branchName: branchName,
                timeout: defaultTimeout
            }
        });
    };
    this.status = function(currentDirectory) {
        return this.sendPayload({
            event: "status",
            data: {
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.statusAllFiles = function(currentDirectory) {
        return this.sendPayload({
            event: "statusAllFiles",
            data: {
                dir: currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.reset = function(currentDirectory, files, maxLength) {
        return this.sendPayload({
            event: "reset",
            data: {
                dir: currentDirectory,
                files: files,
                maxLength: maxLength,
                timeout: defaultTimeout
            }
        });
    };
    this.resetHead = function(currentDirectory, files) {
        return this.sendPayload({
            event: "resetHead",
            data: {
                dir: currentDirectory,
                files: files,
                timeout: defaultTimeout
            }
        });
    };
    this.resetHard = function(currentDirectory) {
        return this.sendPayload({
            event: "resetHard",
            data: {
                dir   : currentDirectory,
                timeout: defaultTimeout
            }
        });
    };
    this.resetSoft = function(currentDirectory, branchName){
        return this.sendPayload({
            event: "resetSoft",
            data: {
                dir : currentDirectory,
                branchName  : branchName,
                timeout: defaultTimeout
            }
        });
    };
    this.copy = function(src, dest, ignoredPaths) {
        return this.sendPayload({
            event: "copy",
            data: {
                dir: {
                    src: src,
                    dest: dest
                },
                ignoredPaths: ignoredPaths,
                timeout: 600000
            }
        });
    };
    this.diffDirectories = function(path, pathToDiff, ignoredPaths) {
        return this.sendPayload({
            event: "diffDirectories",
            data: {
                dir: path,
                dirToDiff: pathToDiff,
                ignoredPaths: ignoredPaths,
                timeout: defaultTimeout
            }
        });
    };
    this.remove = function(path) {
        return this.sendPayload({
            event: "remove",
            data: {
                path: path,
                timeout: 600000
            }
        });
    };
    this.fileExists = function(path, file) {
        return this.sendPayload({
            event: "fileExists",
            data: {
                path: path,
                file: file,
                timeout: defaultTimeout
            }
        });
    };
    this.ls = function(path) {
        return this.sendPayload({
            event: "ls",
            data: {
                path: path,
                timeout: defaultTimeout
            }
        });
    };
    this.lsWithFilterDeploy = function(path, filter, limitFolders) {
        return this.sendPayload({
            event: "lsWithFilterDeploy",
            data: {
                path: path,
                filter: filter,
                limitFolders: limitFolders,
                timeout: defaultTimeout
            }
        });
    };
    this.createFile = function(path, fileName, fileContents, isIncrement) {
        return this.sendPayload({
            event: "createFile",
            data: {
                path: path,
                fileName: fileName,
                fileContents: fileContents,
                isIncrement: isIncrement || false,
                timeout: defaultTimeout
            }
        });
    };
    this.createCompressedArchive = function(path, archiveName, files) {
        return this.sendPayload({
            event: "createCompressedArchive",
            data: {
                path: path,
                archiveName: archiveName,
                files: files,
                timeout: 600000
            }
        });
    };
    this.lsRemote = function(path, url) {
        return this.sendPayload({
            event: "lsRemote",
            data: {
                path: path,
                url: url,
                timeout: defaultTimeout
            }
        });
    };
    this.currentGitRemote = function(path) {
        return this.sendPayload({
            event: "currentGitRemote",
            data: {
                path: path,
                timeout: defaultTimeout
            }
        });
    };
    this.currentGitDiff  = function(currentDirectory ,branchBase, branchToCompare, fileName) {
        return this.sendPayload({
            event: "diff",
            data: {
                branchBase: branchBase,
                branchToCompare: branchToCompare,
                dir: currentDirectory,
                fileName: fileName,
                timeout: defaultTimeout
            }
        });
    };
    this.wget = function(path, url, fileName) {
        return this.sendPayload({
            event: "wget",
            data: {
                path: path,
                url: url,
                fileName: fileName,
                timeout: 3600000
            }
        });
    };
    this.decompressArchive = function(path, archiveName) {
        return this.sendPayload({
            event: "decompressArchive",
            data: {
                path: path,
                archiveName: archiveName,
                timeout: 600000
            }
        });
    };
    this.move = function(src, dest) {
        return this.sendPayload({
            event: "move",
            data: {
                dir: {
                    src: src,
                    dest: dest
                },
                timeout: defaultTimeout
            }
        });
    };
    this.cat = function(path, fileName) {
        return this.sendPayload({
            event: "cat",
            data: {
                path: path,
                fileName: fileName,
                timeout: defaultTimeout
            }
        });
    };
    this.validateExecuteInWindows = function() {
        return this.sendPayload({
            event: "validateExecuteInWindows",
            data: {
                windows: true,
                timeout: defaultTimeout
            }
        });
    };
    this.isOpen = function() {
        return webSocket.isOpen();
    };
    this.error = function(fn) {
        this.catch(fn);
        return this;
    };
    var log = function(name, obj) {
        //console.log(name, obj, (new Date().getTime()));
    };
}
Configuration(function(ContextRegister) {
    ContextRegister.register('GitQueue', GitQueue);
});